/*
  Hangul Drawing routine used in X

  version 0.2

  1996.9 Choi Jun-Ho(junker@vishnu.snu.ac.kr)
  Narae, CS Dept, SNU.
*/

#include <X11/Xlib.h>
#include <stdlib.h>
#include <string.h>
#include <stdio.h>

#include "nxdraw.h"

#define ABS(x)    ((x)>0?(x):-(x))
/* utils */
static int _NXConvEUC_KR7();
static int _EUC_KR2WC();
static XFontStruct* _NXQueryMatchingFont();

/* globals */
static NXFont* hfont;
static NXFont defaultfont;
Display* NX_Display=NULL;

/* NXOpenDisplay()
   
   Equivalent to XOpenDisplay().
   Initialize Dynamic Font Loading Environment.
   Not load fonts actually.

   Recommended to all immediately after XOpenDisplay() calls.
   */
Display* NXOpenDisplay(display_name)
char* display_name;
{
  if ((NX_Display=XOpenDisplay(display_name))==NULL) return NULL; /* cannot init */

  hfont=NULL;

  /* setup default font */
  defaultfont.font=XLoadQueryFont(NX_Display, NX_DEFAULTFONT);

  return NX_Display;
}

/* NXRegisterFontByName()

   Add <fontname> to font list.
   First font registered is used as default font if appropriate font
   doesn't exist.

   This call load font using XLoadQueryFont() calls to get ascent
   metrics.
   */
void NXRegisterFontByName(fontname)
char* fontname;
{
  NXFont* backup;

  if (NX_Display==NULL) return; /* cannot do without display */

  /* add font to chain */
  backup=hfont;
  hfont=(NXFont*)malloc(sizeof(NXFont));
  hfont->fontname=strdup(fontname);
  hfont->font=XLoadQueryFont(NX_Display, fontname);
  hfont->ascent=hfont->font->ascent;
  hfont->fid=hfont->font->fid;
  hfont->next=backup;
  
#ifdef DEBUG
  fprintf(stderr, "Font %s added\n", fontname);
#endif

  return;
}

/* NXDrawString()

   Equivalent to XDrawString() 8bit char version.
   If Korean character appears, find appropriate Korean font and
   display on this.
   */
void NXDrawString(display, d, gc, x, y, str, len)
Display *display;
Drawable d;
GC gc;
int x, y;
char *str;
int len;
{
  int n;
  int clen;
  char *p, *cbuf;
  char cbuf_tmp[MAX_LINE];
  XChar2b buf[MAX_LINE];
  XFontStruct *efont, *hfont;
  Font efid;
  GC hgc;
  XGCValues gcv;

  /* create GC for hangul font */
  hgc=XCreateGC(display, d, 0, (XGCValues *) NULL);
  XCopyGC(display, gc, ~(0L), hgc);

  /* get english font info */
  XGetGCValues(display, gc, GCFont, &gcv);
  efid=gcv.font;
  efont=XQueryFont(display, efid);

  /* get hangul font according to english font */
  hfont=_NXQueryMatchingFont(display, efont->ascent);

  XSetFont(display, hgc, hfont->fid);
  
  /* convert chars if they are in 7-bit mode */
  clen=_NXConvEUC_KR7(str, cbuf_tmp, len); 

  cbuf=cbuf_tmp;
  while (clen > 0) {
    if (*cbuf & 0x80) { /* hangul chars */
      for (p = cbuf; clen > 0 && (*p & 0x80); p += 2, clen -= 2)
        ;
      n = _EUC_KR2WC(cbuf, buf, p-cbuf);
      XDrawString16(display, d, hgc, x, y, buf, n);
      x+=XTextWidth16(hfont, buf, n);
    }
    else { /* english chars */
      for (p = cbuf; clen > 0 && !(*p & 0x80); p++, clen--)
        ;
      XDrawString(display, d, gc, x, y, cbuf, p-cbuf);
      x+=XTextWidth(efont, cbuf, p-cbuf);
    }
    cbuf = p;
  }

  /* clean up */
  XFreeGC(display, hgc);
}

/* NXTextWidth()

   Equivalent to XTextWidth()
   */
int NXTextWidth(efont, str, len)
XFontStruct *efont;
char *str;
int len;
{
  int n;
  int clen;
  int width;
  char *p, *cbuf;
  char cbuf_tmp[MAX_LINE];
  XChar2b buf[MAX_LINE];
  XFontStruct *hfont;

  /* get hangul font according to english font */
  hfont=_NXQueryMatchingFont(NX_Display, efont->ascent);

  /* convert chars if they are in 7-bit mode */
  clen=_NXConvEUC_KR7(str, cbuf_tmp, len); 

  cbuf=cbuf_tmp;
  width=0;
  while (clen > 0) {
    if (*cbuf & 0x80) { /* hangul chars */
      for (p = cbuf; clen > 0 && (*p & 0x80); p += 2, clen -= 2)
        ;
      n = _EUC_KR2WC(cbuf, buf, p-cbuf);
#ifdef DEBUG
      fprintf(stderr, "(hfont, buf, n)=(%p, %s, %d)\n", hfont, buf, n);
#endif
      width+=XTextWidth16(hfont, buf, n);
    }
    else { /* english chars */
      for (p = cbuf; clen > 0 && !(*p & 0x80); p++, clen--)
        ;
      width+=XTextWidth(efont, cbuf, p-cbuf);
    }
    cbuf = p;
  }
  
  return width;
}

/* NXCloseDisplay()

   Equivalent to XCloseDisplay().
   Clear all loading fonts.
   Recommended to use immediately before XCloseDisplay()
   */
void NXCloseDisplay(display)
Display *display;
{
  NXFont* backup;

  if (hfont==NULL){
    XCloseDisplay(display);
    return;
  }

  /* delete default font */
  if (defaultfont.font!=NULL)
    XFreeFont(NX_Display, defaultfont.font);  

  backup=hfont;
  while(backup!=NULL){
#ifdef DEBUG
    fprintf(stderr, "Free Font %s\n", backup->fontname);
#endif
    XFreeFont(NX_Display, backup->font);
    free(backup->fontname);
    backup=backup->next;
  }

  XCloseDisplay(display); 
  return;
}


/****** utils *******/

/* _EUC_KR2WC()
   
   convert EUC-KR(KSC-5601) char string to XChar2b struct.
   corresfonding XLFD *-KSC5601.1987-0
   */
int _EUC_KR2WC(str, buf, n)
char* str;
XChar2b* buf;
int n;
{
  int sp, xcp;

  for (sp = 0, xcp = 0; sp < n;) {
    /* in case of 8bit 2b-char */
    buf[xcp].byte1 = str[sp] - 0x80;
    buf[xcp].byte2 = str[sp+1] - 0x80;
    sp += 2, xcp++;
  }
  return xcp;
}

/* _NXConvEUC_KR7()

    convert 7-bit KSC-5601 code to 8-bit KSC-5601 code
      ESC-$-(-C (0x1b 0x24 0x28 0x43) turns on 7-bit mode
      ESC-(-B (0x1b 0x28 0x42) turns off 7-bit mode
      in 7-bit mode, char must be in 0x21~0x7e
      MSB of char in 7-bit mode is set
*/
int _NXConvEUC_KR7(string, deststr, len)
char* string;
char* deststr;
int len;
{
  int i, j;
  int is_7bit;

  i=0; j=0; is_7bit=0;

  while(i<len){
    if (string[i]==0x1b){
      if (i+4<len &&
          string[i+1]==0x24 &&
          string[i+2]==0x28 &&
          string[i+3]==0x43){
        is_7bit=1;  /* turns on 7bit mode */
        i+=4;       /* jump to next char */
        continue;
      }
      if (i+3<len &&
          string[i+1]==0x28 &&
          string[i+2]==0x42){
        is_7bit=0; /* turns off 7bit mode */
        i+=3;
        continue;
      }
      /* normal ESC sequence */
    }
    if (is_7bit==1 && (string[i]>=0x21 && string[i]<=0x7e))
      /* set MSB */
      deststr[j++]=string[i] | 0x80;
    else
      deststr[j++]=string[i];
    i++;
  }

  return j;
}


/* _NXQueryMatchingFont()

   following to <ascent>, find Korean font and returns XFontStruct*.
   */
XFontStruct* _NXQueryMatchingFont(display, asc)
Display* display;
int asc;
{
  NXFont *curfont, *selfont;
  int min=255;

#ifdef DEBUG
  fprintf(stderr, "NXGetFID: Ascent %d..", asc);
#endif

  /* set default font */
  selfont=&defaultfont;

  /* select best approxicated size font */
  curfont=hfont;
  while(curfont!=NULL){
#ifdef DEBUG
    fprintf(stderr, "%d..", curfont->ascent);
#endif
    if (ABS((asc-1)-curfont->ascent)<=min){
      min=ABS(asc-curfont->ascent);
      selfont=curfont;
#ifdef DEBUG
      fprintf(stderr, "(%d,%d)M%d", curfont->ascent, asc, min);
#endif
    }
    curfont=curfont->next;
  }

  /* error */
  if (selfont==NULL)
    return NULL;

#ifdef DEBUG
  fprintf(stderr, "Font %s selected\n", selfont->fontname);
#endif
  
  return selfont->font;
}
